/* from iabg ipv6_ipsec-main.c -mk */
/*
 * IPSECv6 code
 * Copyright (C) 2000 Stefan Schlott
 * Copyright (C) 2001 Florian Heissenhuber
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/ipv6.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/smp.h>
#include <linux/list.h>

#include <asm/uaccess.h>

#include <net/ipv6.h>
#include <net/sadb.h>
#include <net/spd.h>

#include <linux/ipsec6.h>

#include <net/ipsec6_algos.h>

#include <net/ip6_route.h>
#include <net/addrconf.h>
#include <net/snmp.h>  
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/ipsec.h>
#include <net/ipsec6_utils.h>
#include <linux/ip6packettools.h>
#include "pfkey.h"
#if 0
#include "sadb6.h"
#endif
#include "algos.h"
#include "enc-cbc3des.h"
#include "libdes/des.h"


/* DEBUG *******************************************/
#ifdef CONFIG_IPV6_IPSEC_DEBUG 
#define ipsec6_pr(fmt, arg...) printk(KERN_DEBUG fmt,##arg)
#else
#define ipsec6_pr(fmt,arg...) \
        do { } while (0)
#endif /* CONFIG_IPV6_IPSEC_DEBUG */
/**************************************************/

/* Greatest common divisor */
inline int gcd(int a, int b) {
        int c;
        if (a<b) {
                c=a;
                a=b;
                b=c;
        }
        while (b!=0) {
                c=a;
                a=b;
                b=c%a;
        }
        return a;
}

/* Least common multiple */
inline int lcm(int a, int b) {
        if (a==0 || b==0) return 0;
        return (a*b)/gcd(a,b);
}

char* build_whole_ip6_packet(struct sk_buff *skb, const void *data, 
        unsigned length, inet_getfrag_t getfrag, int *packetlen) 
{
        int totallen;
        char* result;
        
        if (!skb) return NULL;
        if (!data || !getfrag) length=0;
        totallen = ntohs(skb->nh.ipv6h->payload_len) + sizeof(struct ipv6hdr);
        if (totallen != skb->len + length) {
                if (net_ratelimit()) {
                        printk(KERN_WARNING "build_whole_ip6_packet: Something is very wrong...\n");
                        printk(KERN_DEBUG " data length = %u\n",length);
                        printk(KERN_DEBUG " skb->len = %u\n",skb->len);
                        printk(KERN_DEBUG " payload_len = %u\n",ntohs(skb->nh.ipv6h->payload_len));
                }
                return NULL;
        }
        result = kmalloc(totallen,GFP_ATOMIC);
        if (packetlen) *packetlen = totallen;
        if (!result) return NULL;
        memcpy(result,skb->nh.ipv6h,skb->len);
        if (length>0) {
                struct in6_addr *addr;

                addr=&(skb->nh.ipv6h->saddr);
                getfrag(data,addr,&result[skb->len],0,length);
        }
        return result;
}


/* Set all mutable/unpredictable fields to zero. */
void prepare_hash_in_calculation(struct ip6_packet_refs refs) 
{
        if (!refs.ipv6hdr) {
#ifdef IPSEC6_DEBUG
		ipsec6_pr("prepare_hash_in_calculation: No ipv6hdr given!\n");
#endif /* IPSEC6_DEBUG */
                return;
        }
        /* Main header */
        refs.ipv6hdr->tclass1=0;
        refs.ipv6hdr->tclass2_flow[0]=0;
        refs.ipv6hdr->tclass2_flow[1]=0;
        refs.ipv6hdr->tclass2_flow[2]=0;
        refs.ipv6hdr->hop_limit=0;
        /* Mutable/unpredictable Option headers */
        /* AH header */
        if (refs.authhdr) {
                int len = (refs.authhdr->hdrlen*4)-4;
                memset(&refs.authhdr->auth_data[0],0,len);
        }
}

/* Set all mutable/predictable fields to the destination state, and all
   mutable/unpredictable fields to zero. */
void prepare_hash_out_calculation(struct ip6_packet_refs refs) 
{
        prepare_hash_in_calculation(refs);
        /* destination address with routing header ext. */
        /* Routing header */
        /* Mutable/predictable Options */
}


void do_ipsec6_out_calc_ah(struct sock *sk, const void *data,
        unsigned length, inet_getfrag_t getfrag,
        struct sk_buff *skb, struct ipv6_auth_hdr *authhdr, void *ipsec_ptr)
{

        struct sa_index *sa_idx = (struct sa_index*)ipsec_ptr;
        struct ipsec_hash_t *algo_hash;
        char* packet;
        int packetlen,authhdroffset;
        struct ip6_packet_refs refs;
        __u8* authdata;

	ipsec6_pr("do_ipsec6_calc_ah called.\n");

        if (!(sa_idx && sa_idx->sa)) {
		ipsec6_pr("do_ipsec6_out_calc_ah: ipsec6_ptr/sa_transp_ah is NULL!\n");
                return;
        }
        algo_hash = ipsec_get_hash(sa_idx->sa->auth_algo.algo);
        if (!algo_hash) {
		ipsec6_pr("do_ipsec6_out_calc_ah: Hash algorithm %d not present.\n",
			  sa_idx->sa->auth_algo.algo);
                return;
        }
        if (!data) length=0;
        if (!skb) {
		ipsec6_pr("do_ipsec6_out_cal_ah: skb is NULL!\n");
                return;
        }
        /* Handle sequence number */
        sa_idx->sa->replay_window.seq_num++;
        /* TODO: Because 2 packets increment seq_num at the same time,
                seq_num maybe increamented by atomic_inc */

        /* Build packet for calculation */
        packet = build_whole_ip6_packet(skb,data,length,getfrag,&packetlen);
        if (!packet) return;
        analyze_ip6_packet(packet,&refs);
        /* If authhdr not given: Go find it */
        if (!authhdr) {
                authhdroffset = ((char*)refs.authhdr)-packet;
                authhdr = (struct ipv6_auth_hdr*)(((char*)skb->nh.ipv6h)+authhdroffset);
        }
        /* Set authhdr values */
        /* authhdr->spi = htonl(sa_idx->spi); */
	ipsec6_pr("do_ipsec6_out_calc_ah: sa_idx->spi is 0x%x\n", sa_idx->spi);
	authhdr->spi = sa_idx->spi; /* -mk */
        authhdr->seq_no = htonl(sa_idx->sa->replay_window.seq_num);
	ipsec6_pr("do_ipsec6_out_calc_ah: authhdr->spi is 0x%x\n", authhdr->spi);
        refs.authhdr->spi = authhdr->spi;
        refs.authhdr->seq_no = authhdr->seq_no;
        /* Finally: Calculate and copy checksum */
        prepare_hash_out_calculation(refs);

        authdata=kmalloc(algo_hash->hash_size, GFP_ATOMIC);
        if (!authdata) {
                if (net_ratelimit())
                        printk(KERN_WARNING "do_ipsec6_out_calc_ah: Out of memory.\n");
                return;
        }
        algo_hash->calc_hash(packet,packetlen,
                sa_idx->sa->auth_algo.key,authdata);
        memcpy(&authhdr->auth_data[0],authdata,algo_hash->hash_size);
	ipsec6_pr( "do_ipsec6_sout_calc_ah: Packet structure:\n");
	print_ip6packet(refs);
	ipsec6_pr("do_ipsec6_out_calc: End packet info (out).\n");
        kfree(authdata);
        kfree(packet); 
}


int do_ipsec6_out_req_header_size(int action, struct sock *sk, void *ipsec_ptr) 
{
        int result = 0;
        struct sa_index *sa_idx = (struct sa_index*)ipsec_ptr;
        struct ipsec_hash_t *hash;
        struct ipsec_enc_t *enc;

	ipsec6_pr("do_ipsec6_req_header_size called.\n");

        if (!sa_idx) return 0;

        if ((action & IPSEC_ACTION_AUTH) && sa_idx->sa) {
                hash=ipsec_get_hash(sa_idx->sa->auth_algo.algo);
                if (hash) {
                        result+=sizeof(struct ipv6_auth_hdr)-4;
                        result+=hash->hash_size;
                        /* 64 bit alignment */
                        if ((hash->hash_size-4)%8 !=0 )
                                result+=8-((hash->hash_size-4)%8);
                }
        }
        if ((action & IPSEC_ACTION_ESP) && sa_idx->sa) {
                hash=ipsec_get_hash(sa_idx->sa->esp_algo.algo);
                enc=ipsec_get_enc(sa_idx->sa->auth_algo.algo);
                result+=sizeof(struct ipv6_esp_hdr)-8;
                result+=hash->hash_size;
                if (hash->hash_size%8 !=0 ) /* 64 bit alignment */
                        result+=8-(hash->hash_size%8);
                result+=enc->iv_size;
                /* if (enc->block_size<8) result+=8; else
                        result+=enc->block_size; */
                result+=lcm(enc->block_size,8); /* Block size and 64 bit alignment */
        }
	ipsec6_pr("do_ipsec6_req_header_size: Calculated size is %d.\n", result);
        return result;
}


struct ipv6_txoptions * do_ipsec6_out_insert_ah(struct sock *sk,
	const void *data, unsigned length, struct ipv6_txoptions *opt,
	void *ipsec_ptr) {
	struct ipv6_txoptions *dummyopt = NULL;
	struct ipv6_auth_hdr *dummyauthhdr = NULL;
	struct sa_index *sa_idx = (struct sa_index*)ipsec_ptr;
	struct ipsec_hash_t *hash;
	int ahlen;
	
	ipsec6_pr("do_ipsec6_out_insert_ah called.\n");
	if (!(sa_idx && sa_idx->sa)) {
		if (net_ratelimit())
			printk(KERN_INFO"do_ipsec6_out_insert_ah: ipsec6_ptr/sa_transp_ah missing.\n");
		return NULL;
	}
	hash = ipsec_get_hash(sa_idx->sa->auth_algo.algo);
	if (!hash) {
		if (net_ratelimit())
			printk(KERN_INFO "do_ipsec6_out_insert_ah: Hash algorithm %d not present.\n",
				sa_idx->sa->auth_algo.algo);
		return NULL;
	} 
	ahlen = sizeof(struct ipv6_auth_hdr) - 4 + hash->hash_size;
	if ((hash->hash_size-4)%8 !=0 )
		ahlen+=8-((hash->hash_size-4)%8); /* 64 bit alignment */

	ipsec6_pr("do_ipsec6_out_insert_ah: ahlen=%d hash_size=%d\n", ahlen,hash->hash_size);

	if (opt) {
		dummyopt = kmalloc(opt->tot_len + ahlen, GFP_ATOMIC);
		if (!dummyopt) {
			if (net_ratelimit())
				printk(KERN_WARNING "Couldn't allocate AH header - out of memory.\n");
			return NULL;
		}
		memcpy(dummyopt, opt, opt->tot_len);
		/* Oh Jesus... please adjust pointers of new dummyopt! (Todo) */
		dummyauthhdr = (struct ipv6_auth_hdr*) (((__u8*)dummyopt) + opt->tot_len);
	} else {
		dummyopt = kmalloc(sizeof(struct ipv6_txoptions) + ahlen, GFP_ATOMIC);
		if (!dummyopt) {
			if (net_ratelimit())
				printk(KERN_WARNING "Couldn't allocate AH header - out of memory.\n");
			return NULL;
		}
		memset(dummyopt, 0, sizeof(struct ipv6_txoptions));
		dummyauthhdr = (struct ipv6_auth_hdr*) (((__u8*)dummyopt) + sizeof(struct ipv6_txoptions));
	}
	memset(dummyauthhdr, 0, ahlen);
	dummyopt->auth = (struct ipv6_opt_hdr*) dummyauthhdr;
	dummyopt->tot_len += ahlen;
	dummyopt->opt_flen += ahlen;
	dummyauthhdr->hdrlen = (ahlen/4) - 2;
	ipsec6_pr("do_ipsec6_insert_ah: Dummy auth header inserted in opt\n"); 
	return dummyopt;
}


